﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;

namespace SoutezniAplikace
{
    public partial class Form1 : Form
    {
        public string HostPlusPort;
        public string GameID;
        public List<List<Tile>> Map;
        public Point LerakPosition;
        public Direction LerakDirection;
        public Point TreasuryPosition;
        public string PossibleMovements;
        public bool GameInProgress = false;

        public Form1()
        {
            InitializeComponent();
        }

        // Založí novou hru
        private void bNewGame_Click(object sender, EventArgs e)
        {
            
            HostPlusPort = "http://" + this.tbHost.Text + ":" + this.nPort.Value.ToString() + "/";
            StartGame();
            RefreshMap();
            this.bSendRound.Enabled = true;
        }

        // Aktualizuje zobrazení mapy
        public void RefreshMap()
        {
            this.pictureBox1.Invalidate();
            this.pictureBox1.Update();

            this.lblMoznePohyby.Text = "Možné pohyby: " + PossibleMovements;
        }

        public void StartGame()
        {
            // Pošle požadavak serveru
            string[] responses = Request("start").Split(new string[] { "\r\n" }, StringSplitOptions.None);
            GameID = responses[0];
            demi_steps = new List<Point>();
            this.tbPohyby.Text = "";

            // Vytvoří mapu v paměti
            Map = new List<List<Tile>>();
            for (int x = 0; x < 20; x++)
            {
                Map.Add(new List<Tile>());
                for (int y = 0; y < 20; y++)
                {
                    Map[x].Add(new Tile(x,y));
                }
            }

            // Zkonvertuje text z odpovědi serveru do paměťové reprezentace mapy
            for (int y = 0; y < 20; y++)
            {
                for (int x = 0; x < 20; x++)
                {
                    Map[x][y].Type = TileTypeFromChar(responses[y+1][x]);
                    Map[x][y].Char = responses[y+1][x];
                    if (Map[x][y].Type == TileType.Treasury)
                    {
                        TreasuryPosition = new Point(x, y);
                    }
                }
            }

            // Zjistí pozici Leraka.
            string[] lerak = responses[21].Split(' ');
            LerakPosition = new Point(Int32.Parse(lerak[0]) - 1, Int32.Parse(lerak[1]) - 1);
            LerakDirection = DirectionFromChar(lerak[2][0]);
            PossibleMovements = responses[22];
            GameInProgress = true;
        }

        // Pošle požadavak HTTP.
        // V souladu se zadáním očekávám, že server bude vždy běžet a vracet korektní odpovědi, proto žádné ošetření výjimek.
        public string Request(string query)
        {
            WebClient wc = new WebClient();
            return wc.DownloadString(HostPlusPort + query);
          
        }

        // Zkonvertuje znak označující políčko na jeho reprezentaci v paměti
        public TileType TileTypeFromChar(char c)
        {
            switch (c)
            {
                case '.': return TileType.Empty;
                case 'X': return TileType.Wall;
                case '#': return TileType.Pit;
                case '$': return TileType.Treasury;
                case '^': return TileType.GoUp;
                case 'v': return TileType.GoBottom;
                case '<': return TileType.GoLeft;
                case '>': return TileType.GoRight;
            }
            return TileType.Empty;
        }

        // Převede znak označující směr na hezkou konstantu
        public Direction DirectionFromChar(char c)
        {
            switch (c)
            {
                case 'L': return Direction.Left;
                case 'R': return Direction.Right;
                case 'N': return Direction.Up;
                case 'D': return Direction.Down;
            }
            return Direction.Up;
        }

        // Převede směr zpět na znak, aby mohl být odeslán serveru
        public char CharFromDirection(Direction d)
        {
            switch (d)
            {
                case Direction.Left: return 'L';
                case Direction.Right: return 'R';
                case Direction.Up: return 'N';
                case Direction.Down: return 'D';
            }
            return 'N';
        }

        // Odešle požadavak na nové kolo.
        private void bSendRound_Click(object sender, EventArgs e)
        {
            if (this.tbPohyby.Text == "") return;
            string[] responses = Request("kolo?id=" + GameID + "&pohyby=" + this.tbPohyby.Text).Split(new string[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
            
            // Jednořádková odpověď znamená konec hry.
            if (responses.Length == 1)
            {
                MessageBox.Show(responses[0], "Konec hry");
                GameInProgress = false;
                this.bSendRound.Enabled = false;
            }
            else
            {
                // Jinak se změní Lerakova pozice a směr, a možné pohyby.
                string[] lerak = responses[21].Split(' ');
                LerakPosition = new Point(Int32.Parse(lerak[0]) - 1, Int32.Parse(lerak[1]) - 1);
                LerakDirection = DirectionFromChar(lerak[2][0]);
                PossibleMovements = responses[22];
                demi_steps = new List<Point>();
                this.tbPohyby.Text = "";
                this.lblMoznePohyby.Text = "Možné pohyby: " + PossibleMovements;
                RefreshMap();
            }
        }

        // Velikost políčka na obrazovce
        int TILESIZE = 20;

        // Vykreslí mapu
        private void pictureBox1_Paint(object sender, PaintEventArgs e)
        {
            Rectangle all_map = new Rectangle(0, 0, TILESIZE * 20, TILESIZE * 20);
            e.Graphics.FillRectangle(Brushes.White, all_map);
            Font font = new Font(FontFamily.GenericSansSerif, 8);


            if (GameInProgress)
            {
                // Políčka
                for (int x = 0; x < 20; x++)
                {
                    for (int y = 0; y < 20; y++)
                    {
                        Brush brush = Brushes.White;
                        Char cPasy = '0';

                        switch (Map[x][y].Type)
                        {
                            case TileType.Wall: brush = Brushes.MidnightBlue; break;
                            case TileType.Pit: brush = Brushes.Brown; break;
                            case TileType.Treasury: brush = Brushes.Yellow; break;
                            case TileType.GoBottom: brush = Brushes.Pink; cPasy = 'v'; break;
                            case TileType.GoLeft: brush = Brushes.Pink; cPasy = '<'; break;
                            case TileType.GoRight: brush = Brushes.Pink; cPasy = '>'; break;
                            case TileType.GoUp: brush = Brushes.Pink; cPasy = '^'; break;
                        }

                        e.Graphics.FillRectangle(brush, new Rectangle(x * TILESIZE, y * TILESIZE, TILESIZE, TILESIZE));
                        // Pasy
                        if (cPasy != '0')
                        {
                            e.Graphics.DrawString(cPasy.ToString(), font, Brushes.Black, new PointF(x * TILESIZE, y * TILESIZE));
                        }
                    }



                }

                // Robot Karel
                e.Graphics.FillRectangle(Brushes.Lime, new Rectangle(LerakPosition.X * TILESIZE, LerakPosition.Y * TILESIZE, TILESIZE, TILESIZE));
                e.Graphics.DrawString(CharFromDirection(LerakDirection).ToString(), font, Brushes.Black, new PointF(LerakPosition.X * TILESIZE, LerakPosition.Y * TILESIZE));

                // Čára předpovídající, kam se Lerak dostane, pokud by hráčem vybrané příkazy byly provedeny
                if (demi_steps.Count != 0)
                {
                    Point old = LerakPosition;
                    foreach (Point p in demi_steps)
                    {
                        e.Graphics.DrawLine(Pens.Black, new Point(old.X * TILESIZE + TILESIZE / 2, old.Y * TILESIZE + TILESIZE / 2), new Point(p.X * TILESIZE + TILESIZE / 2, p.Y * TILESIZE + TILESIZE / 2));
                        old = p;
                    }
                }

                // Čáry kolem políček
                for (int x = 0; x < 20; x++)
                {
                    e.Graphics.DrawLine(Pens.Gray, x * TILESIZE, 0, x * TILESIZE, 20 * TILESIZE);
                }
                for (int y = 0; y < 20; y++)
                {
                    e.Graphics.DrawLine(Pens.Gray, 0, y * TILESIZE, TILESIZE * 20, y * TILESIZE);
                }
            }
            e.Graphics.DrawRectangle(Pens.Black, all_map);
        }

        // Aktualizuje předpověď pozice Leraka
        private void tbPohyby_TextChanged(object sender, EventArgs e)
        {
            if (this.bSendRound.Enabled)
            {
                PredictPosition(this.tbPohyby.Text);
                RefreshMap();
            }
        }

        // Jednotlivé kroky Leraka, tak, jak byly předpovězeny
        public List<Point> demi_steps = new List<Point>();

        // Převe soubor pěti pohybů (= příkaz) na výslednou pozici Leraka, a jednotlivé pohyby uloží do demi_steps
        public void PredictPosition(string order)
        {
            demi_steps = new List<Point>();
            Point newPosition = LerakPosition;
            Direction newDirection = LerakDirection;
            foreach (char c in order)
            {
                switch (c)
                {
                    case '1':
                        newPosition = Move(newPosition, newDirection, 1);
                        break;
                    case '2':
                        newPosition = Move(newPosition, newDirection, 2);
                        break;
                    case '3':
                        newPosition = Move(newPosition, newDirection, 3);
                        break;
                    case 'Z':
                        newPosition = Move(newPosition, newDirection, -1);
                        break;
                    case 'L':
                        newDirection = Redirect(newDirection, -1);
                        break;
                    case 'R':
                        newDirection = Redirect(newDirection, 1);
                        break;
                }
                newPosition = Escalators(newPosition);
                demi_steps.Add(newPosition);
            }
            if (newPosition.X == -1 && newPosition.Y == -1)
            {
                // To znamená, že Lerak zemře. 
                demi_steps = new List<Point>();
            }
        }

        // Utilita pro umělou inteligenci. Vypočítá pozici, kam se Lerak dostane po provedení příkazu ORDER
        public Point GetNewPosition(Point oldPos, Direction oldDir, string order, out Direction direction)
        {
            demi_steps = new List<Point>();
            Point newPosition = oldPos;
            Direction newDirection = oldDir;
            foreach (char c in order)
            {
                switch (c)
                {
                    case '1':
                        newPosition = Move(newPosition, newDirection, 1);
                        break;
                    case '2':
                        newPosition = Move(newPosition, newDirection, 2);
                        break;
                    case '3':
                        newPosition = Move(newPosition, newDirection, 3);
                        break;
                    case 'Z':
                        newPosition = Move(newPosition, newDirection, -1);
                        break;
                    case 'L':
                        newDirection = Redirect(newDirection, -1);
                        break;
                    case 'R':
                        newDirection = Redirect(newDirection, 1);
                        break;
                }
                newPosition = Escalators(newPosition);
            }
            direction = newDirection;
            return newPosition;
        }

        // Posune Leraka, stojí-li na pásu
        private Point Escalators(Point oldPosition)
        {
            
            if (oldPosition.X != -1)
            {
                Point newPosition = oldPosition;
                switch (Map[oldPosition.X][oldPosition.Y].Type)
                {
                    case TileType.GoLeft: newPosition = new Point(oldPosition.X - 1, oldPosition.Y); break;
                    case TileType.GoRight: newPosition = new Point(oldPosition.X + 1, oldPosition.Y); break;
                    case TileType.GoUp: newPosition = new Point(oldPosition.X, oldPosition.Y - 1); break;
                    case TileType.GoBottom: newPosition = new Point(oldPosition.X, oldPosition.Y + 1); break;
                }
                if (newPosition == oldPosition) return oldPosition;

                if (newPosition.X < 0 || newPosition.Y < 0 || newPosition.X >= 20 || newPosition.Y >= 20)
                {
                    return new Point(-1, -1);
                }

                switch (Map[newPosition.X][newPosition.Y].Type)
                {
                    case TileType.Wall:
                        return oldPosition;
                    case TileType.Pit:
                        return new Point(-1, -1);
                }
                return newPosition;
            }
            else return oldPosition;
        }

        // Utilita pro pohyby 1,2,3 a Z. Zjistí, kam až se Lerak posune, a jestli přitom zemře.
        // ... nebo narazí na poklad.
        public Point Move(Point oldPosition, Direction direction, int number)
        {
            if (number == -1)
            {
               direction = Redirect(Redirect(direction, -1), -1);
               number = 1;
            }
            Point newPosition = new Point(0, 0);
            for (int i = 0; i < number; i++ )
            {
                if (direction == Direction.Right) newPosition = new Point(oldPosition.X + 1, oldPosition.Y);
                if (direction == Direction.Left) newPosition = new Point(oldPosition.X - 1, oldPosition.Y);
                if (direction == Direction.Up) newPosition = new Point(oldPosition.X, oldPosition.Y - 1);
                if (direction == Direction.Down) newPosition = new Point(oldPosition.X, oldPosition.Y + 1);

                if (newPosition.X < 0 || newPosition.Y < 0 || newPosition.X >= 20 || newPosition.Y >= 20)
                {
                    return new Point(-1, -1);
                }
                switch (Map[newPosition.X][newPosition.Y].Type)
                {
                    case TileType.Wall:
                        newPosition = oldPosition;
                        break;
                    case TileType.Treasury:
                        return newPosition;
                    case TileType.Pit:
                        return new Point(-1, -1);
                }
                oldPosition = newPosition;
            }
            return newPosition;
        }

        // Otočí Leraka doleva, či doprava
        public Direction Redirect(Direction oldDirection, int what_side)
        {
            if (what_side == -1)
            {
                switch (oldDirection)
                {
                    case Direction.Up: return Direction.Left;
                    case Direction.Left: return Direction.Down;
                    case Direction.Down: return Direction.Right;
                    case Direction.Right: return Direction.Up;
                }
            }
            else
            {
                switch (oldDirection)
                {
                    case Direction.Up: return Direction.Right;
                    case Direction.Right: return Direction.Down;
                    case Direction.Down: return Direction.Left;
                    case Direction.Left: return Direction.Up;
                }
            }
            return Direction.Up;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            this.comboBox1.SelectedIndex = 0;
        }

        // Pokusí se vyhledat cestu na dané políčko (s danou orientací!)
        private void pictureBox1_MouseClick(object sender, MouseEventArgs e)
        {
            int targetX = e.X / TILESIZE;
            int targetY = e.Y / TILESIZE;
            Direction dire = Direction.Up;
            switch (this.comboBox1.SelectedIndex)
            {
                case 0: dire = Direction.Left; break;
                case 1: dire = Direction.Right; break;
                case 2: dire = Direction.Up; break;
                case 3: dire = Direction.Down; break;
            }

            if (targetX >= 0 && targetY >= 0 && targetX < 20 && targetY < 20)
            {
                string command = FindPath(LerakPosition, LerakDirection, PossibleMovements, new Point(targetX, targetY), dire);
                this.tbPohyby.Text = command;
            }
        }

        // Multifunkční vyhledávání.
        // Zaprvé slouží k vyhledání cesty ke kliknuté pozici
        // Zadruhé slouží umělé inteligenci k vyhledání cesty k jakékoli pozici
        public string FindPath(Point oldPosition, Direction oldDirection, string possibilities, Point targetPosition, Direction targetDirection)
        {
            // 1. Nejprve vyhledá systém všechna možná políčka, kde Lerak může skončit
            // tak, že vyzkouší všechny možné kombinace příkazů

            List<char> orders = possibilities.ToList<char>();
            List<Searcher> Searchers = new List<Searcher>();
            List<Searcher> Results = new List<Searcher>();
            Searchers.Add(new Searcher(oldPosition, oldDirection, orders, ""));
            bool floodFill = false;
            if (targetPosition.X == -1)
                floodFill = true;
            
            while (Searchers.Count > 0)
            {
                List<Searcher> new_spin_offs = new List<Searcher>();
                for (int i = Searchers.Count - 1; i >= 0; i--)
                {
                    Searcher s = Searchers[i];

                    if (floodFill)
                    {
                        if (s.OrdersDoneLength == 5) Results.Add(s);
                    }
                    else
                    {
                        if (s.OrdersDoneLength == 5)
                        if (s.Position == targetPosition && s.Direction == targetDirection)
                        {
                            return s.OrdersDone;
                        }
                    }

                    foreach (char c in s.OrdersRemaining)
                    {
                        Direction direction = Direction.Down;
                        Point point = GetNewPosition(s.Position, s.Direction, c.ToString(), out direction);

                        if (s.OrdersDoneLength != 5)
                        {
                            if (point.X != -1)
                            {
                                new_spin_offs.Add(new Searcher(point, direction, WithoutSpecificOrder(s.OrdersRemaining, c), s.OrdersDone, c));
                            }
                        }
                    }
                }
                Searchers = new_spin_offs;
                if (floodFill && DateTime.Now > SendAtLatestWhen)
                    break;
            }
            
            // Nyní, pokud jde o umělou inteligenci, využije Dijsktrův algoritmus k nalezení nejkratší cesty do cíle
            // a vybere si políčko, které je cíli nejblíže.
            if (floodFill)
            {
              

                if (this.checkBox1.Checked)
                { 
                    string bestOrder = "";
                    int bestDistanceFromTreasury = Int32.MaxValue;
                    List<Searcher> SelectFew = new List<Searcher>();

                    // Algoritmus se spouští pouze na 100 ukázkových výsledků, abychom šetřili procesorový čas
                    for (int i = 1; i < 100; i++)
                    {
                        if (Results.Count != 0)
                           SelectFew.Add(Results[Results.Count * i / 100]);
                    }

                    foreach (Searcher s in SelectFew)
                    {
                        int distance = Int32.MaxValue;
                        // A*
                        for (int x = 0; x < 20; x++)
                        {
                            for (int y = 0; y < 20; y++)
                            {
                                Map[x][y].DifficultyInReaching = Int32.MaxValue;
                            }
                        }
                        List<Tile> openTiles = new List<Tile>();
                        List<Tile> closedTiles = new List<Tile>();
                        openTiles.Add(Map[s.Position.X][s.Position.Y]);
                        Map[s.Position.X][s.Position.Y].DifficultyInReaching = 0;

                        while (openTiles.Count > 0)
                        {
                            // find lowest open tile
                            int lowestDiff = Int32.MaxValue;
                            Tile activeTile = null;
                            foreach (Tile t in openTiles)
                            {
                                if (t.DifficultyInReaching < lowestDiff)
                                {
                                    lowestDiff = t.DifficultyInReaching;
                                    activeTile = t;
                                }
                            }
                            // close it
                            openTiles.Remove(activeTile);
                            closedTiles.Add(activeTile);
                            // test if it is target
                            if (activeTile.Type == TileType.Treasury)
                            {
                                distance = activeTile.DifficultyInReaching;
                                break;
                            }
                            // proliferate
                            List<Tile> whereToGo = new List<Tile>();
                            if (activeTile.X > 0) whereToGo.Add(Map[activeTile.X - 1][activeTile.Y]);
                            if (activeTile.Y > 0) whereToGo.Add(Map[activeTile.X][activeTile.Y - 1]);
                            if (activeTile.X < 19) whereToGo.Add(Map[activeTile.X + 1][activeTile.Y]);
                            if (activeTile.Y < 19) whereToGo.Add(Map[activeTile.X][activeTile.Y + 1]);

                            foreach (Tile t in whereToGo)
                            {
                                if (!closedTiles.Contains(t) && t.Type != TileType.Wall && t.Type != TileType.Pit)
                                {
                                    if (!openTiles.Contains(t))
                                         openTiles.Add(t);
                                    if (t.DifficultyInReaching > activeTile.DifficultyInReaching)
                                        t.DifficultyInReaching = activeTile.DifficultyInReaching + 1;
                                }
                            }

                        }
                        if (distance < bestDistanceFromTreasury)
                        {
                            bestDistanceFromTreasury = distance;
                            bestOrder = s.OrdersDone;
                        }
                    }
                    return bestOrder;
                }
                else
                {
                 
                   string bestOrder = "";
                    int bestDistanceFromTreasury = Int32.MaxValue;
                    foreach (Searcher s in Results)
                    {
                        int dist_from_treasury = Math.Abs(s.Position.X - TreasuryPosition.X) + Math.Abs(s.Position.Y - TreasuryPosition.Y);
                        if (dist_from_treasury < bestDistanceFromTreasury)
                        {
                            bestDistanceFromTreasury = dist_from_treasury;
                            bestOrder = s.OrdersDone;
                        }
                    }

                    return bestOrder;
                 } 
            }
            return "";
        }

        // Odstraní již použitý pohyb ze seznamu příkazů
        public List<char> WithoutSpecificOrder(List<char> list, char order)
        {
            List<char> n = new List<char>();
            bool yet_to_be_removed = true;
            foreach (char c in list)
            {
                if (c == order && yet_to_be_removed)
                {
                    yet_to_be_removed = false;
                }
                else
                {
                    n.Add(c);
                }
            }
            return n;
        }

        // AI
        
        // Časovač. To aby odešel i horší výsledek, spíše než abychom překročili časový limit
        public DateTime SendAtLatestWhen = DateTime.Now;


        // Spustí umělou inteligenci tak, že vytvoří hru a pak spustí přemýšlení.
        private void button1_Click(object sender, EventArgs e)
        {
            HostPlusPort = "http://" + this.tbHost.Text + ":" + this.nPort.Value.ToString() + "/";
            SendAtLatestWhen = DateTime.Now.AddSeconds(9);
            StartGame();
            if (chRefreshMap.Checked)
                RefreshMap();
            this.bSendRound.Enabled = false;
            this.button1.Enabled = false;
            Think_AI_think();
        }

        // Nalezne nějakou cestu pomocí FindPath, posune tam Leraka a hledá další část cesty.
        // Popř. ukončí hledání, našla-li se chyba, nebo poklad.
        public void Think_AI_think()
        {
            string order = FindPath(LerakPosition, LerakDirection, this.PossibleMovements, new Point(-1,-1), Direction.Down);
            if (order == "")
            {
                MessageBox.Show("Superprogramu se nepodařilo najít řešení včas nebo neexistuje žádná možná cesta.", "Chyba umělé inteligence");
                GameInProgress = false;
                this.bSendRound.Enabled = false;
                this.button1.Enabled = true;
                return;
            }
            SendAtLatestWhen = DateTime.Now.AddSeconds(9);
            string[] responses = Request("kolo?id=" + GameID + "&pohyby=" + order).Split(new string[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
            if (responses.Length == 1)
            {
                MessageBox.Show(responses[0], "Konec hry");
                GameInProgress = false;
                this.bSendRound.Enabled = false;
                this.button1.Enabled = true;
            }
            else
            {
                string[] lerak = responses[21].Split(' ');
                LerakPosition = new Point(Int32.Parse(lerak[0]) - 1, Int32.Parse(lerak[1]) - 1);
                LerakDirection = DirectionFromChar(lerak[2][0]);
                PossibleMovements = responses[22];
                demi_steps = new List<Point>();
                if (chRefreshMap.Checked) RefreshMap();
                Think_AI_think();
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            AboutBox1 about = new AboutBox1();
            about.ShowDialog();
        }

    }
    public class Searcher
    {
        public List<char> OrdersRemaining;
        public string OrdersDone;
        public int OrdersDoneLength;
        public Point Position;
        public Direction Direction;
        public Searcher(Point p, Direction d, List<char> orders, string ordersDoneOld, char newOrderDone = '0')
        {
            Position = p;
            Direction = d;
            OrdersRemaining = orders;
            if (newOrderDone != '0')
            {
                OrdersDone = ordersDoneOld;
                OrdersDone += newOrderDone;
                OrdersDoneLength = OrdersDone.Length;
            }
            
        }
    }
    public enum Direction
    {
        Left,
        Right,
        Up,
        Down
    }
    public class Tile
    {
        public TileType Type;
        public char Char;
        public int DifficultyInReaching;
        public int X;
        public int Y;
        public Tile(int x, int y)
        {
            X = x; Y = y;
        }
    }
    public enum TileType
    {
        Empty,
        Wall,
        Pit,
        GoRight,
        GoLeft,
        GoUp,
        GoBottom,
        Treasury
    }
}
